import pytest
from plette.models import Source
from plette import models

def test_source_from_data():
    s = Source(
        **{
            "name": "devpi",
            "url": "https://$USER:$PASS@mydevpi.localhost",
            "verify_ssl": False,
          }
    )
    assert s.name == "devpi"
    assert s.url == "https://$USER:$PASS@mydevpi.localhost"
    assert s.verify_ssl is False


def test_source_as_data_expanded(monkeypatch):
    monkeypatch.setattr("os.environ", {"USER": "user", "PASS": "pa55"})
    s = Source(
        **{
            "name": "devpi",
            "url": "https://$USER:$PASS@mydevpi.localhost",
            "verify_ssl": False,
        }
    )
    assert s.url_expanded == "https://user:pa55@mydevpi.localhost"


def test_source_as_data_expanded_partial(monkeypatch):
    monkeypatch.setattr("os.environ", {"USER": "user"})
    s = Source(
        **{
            "name": "devpi",
            "url": "https://$USER:$PASS@mydevpi.localhost",
            "verify_ssl": False,
        }
    )
    assert s.url_expanded == "https://user:$PASS@mydevpi.localhost"


def test_validation_error():
    data = {"name": "test", "verify_ssl": 1}

    with pytest.raises(TypeError) as exc_info:
        Source(**data)

    error_message = str(exc_info.value)

    assert "missing 1 required positional argument: 'url'" in error_message

    data["url"]  = "http://localhost:8000"

    with pytest.raises(models.ValidationError) as exc_info:
        Source(**data)

    error_message = str(exc_info.value)


    assert "verify_ssl: must be of boolean type" in error_message



import hashlib
import json

from dataclasses import dataclass, asdict

from typing import Optional

import tomlkit


from .models import (
    BaseModel,
    Hash, Requires, PipfileSection, Pipenv,
    PackageCollection, ScriptCollection, SourceCollection,
    remove_empty_values
)


PIPFILE_SECTIONS = {
    "sources": SourceCollection,
    "packages": PackageCollection,
    "dev-packages": PackageCollection,
    "requires": Requires,
    "scripts": ScriptCollection,
    "pipfile":  PipfileSection,
    "pipenv": Pipenv
}

DEFAULT_SOURCE_TOML = """\
[[source]]
name = "pypi"
url = "https://pypi.org/simple"
verify_ssl = true
"""


@dataclass
class Pipfile(BaseModel):
    """Representation of a Pipfile."""
    sources: SourceCollection
    packages: Optional[PackageCollection] = None
    packages: Optional[PackageCollection] = None
    dev_packages: Optional[PackageCollection] = None
    requires: Optional[Requires] = None
    scripts: Optional[ScriptCollection] = None
    pipfile: Optional[PipfileSection] = None
    pipenv: Optional[Pipenv] = None

    def validate_sources(self, value):
        if isinstance(value, list):
            return SourceCollection(value)
        return SourceCollection(value.value)

    def validate_pipenv(self, value):
        if value is not None:
            return Pipenv(**value)
        return value

    def validate_packages(self, value):
        PackageCollection(value)
        return value

    def to_dict(self):
        data = {
            "_meta": {
                "requires": getattr(self, "requires", {}),
            },
            "default": getattr(self, "packages", {}),
            "develop": getattr(self, "dev-packages", {}),
        }
        data["_meta"].update(asdict(getattr(self, "sources", {})))
        for category, values in self.__dict__.items():
            if category in PIPFILE_SECTIONS or category in (
                    "default", "develop", "pipenv"):
                continue
            data[category] = values
        remove_empty_values(data)
        return data

    def get_hash(self):
        data = self.to_dict()
        content = json.dumps(data, sort_keys=True, separators=(",", ":"))
        if isinstance(content, str):
            content = content.encode("utf-8")
        return Hash.from_hash(hashlib.sha256(content))

    @classmethod
    def load(cls, f, encoding=None):
        content = f.read()
        if encoding is not None:
            content = content.decode(encoding)
        data = tomlkit.loads(content)
        if "source" not in data:
            # HACK: There is no good way to prepend a section to an existing
            # TOML document, but there's no good way to copy non-structural
            # content from one TOML document to another either. Modify the
            # TOML content directly, and load the new in-memory document.
            sep = "" if content.startswith("\n") else "\n"
            content = DEFAULT_SOURCE_TOML + sep + content
        data = tomlkit.loads(content)
        data["sources"] = data.pop("source")
        packages_sections = {}
        data_sections = list(data.keys())
        for k in data_sections:
            if k not in cls.__dataclass_fields__:
                packages_sections[k] = data.pop(k)

        inst = cls(**data)
        if packages_sections:
            for k, v in packages_sections.items():
                setattr(inst, k, PackageCollection(v))
        return inst

    @property
    def source(self):
        return self.sources

    def dump(self, f, encoding=None):
        data = self.to_dict()
        new_data = {}
        metadata = data.pop("_meta")
        new_data["source"] = metadata.pop("sources")
        new_data["packages"] = data.pop("default")
        new_data.update(data)
        content = tomlkit.dumps(new_data)

        if encoding is not None:
            content = content.encode(encoding)
        f.write(content)



# pylint: disable=missing-module-docstring,missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=no-member

import dataclasses
import json
import numbers

import collections.abc as collections_abc

from dataclasses import dataclass, field, asdict
from typing import Optional

from .models import BaseModel, Meta, PackageCollection, Package, remove_empty_values

PIPFILE_SPEC_CURRENT = 6


def flatten_versions(d):
    copy = {}
    # Iterate over a copy of the dictionary
    for key, value in d.items():
        # If the key is "version", replace the key with the value
        copy[key] = value["version"]
    return copy


class DCJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            o = dataclasses.asdict(o)
            if "_meta" in o:
                o["_meta"]["pipfile-spec"] = o["_meta"].pop("pipfile_spec")
                o["_meta"]["hash"] = {o["_meta"]["hash"]["name"]: o["_meta"]["hash"]["value"]}
                o["_meta"]["sources"] = o["_meta"]["sources"].pop("sources")

            remove_empty_values(o)

            for section in ["default", "develop"]:
                try:
                    o[section] = flatten_versions(o[section])
                except KeyError:
                    continue
            # add silly default values
            if "develop" not in o:
                o["develop"] = {}
            if "requires" not in o["_meta"]:
                o["_meta"]["requires"] = {}
            return o
        return super().default(o)


def _copy_jsonsafe(value):
    """Deep-copy a value into JSON-safe types.
    """
    if isinstance(value, (str, numbers.Number)):
        return value
    if isinstance(value, collections_abc.Mapping):
        return {str(k): _copy_jsonsafe(v) for k, v in value.items()}
    if isinstance(value, collections_abc.Iterable):
        return [_copy_jsonsafe(v) for v in value]
    if value is None:   # This doesn't happen often for us.
        return None
    return str(value)


@dataclass
class Lockfile(BaseModel):
    """Representation of a Pipfile.lock."""

    _meta: Optional[Meta]
    default: Optional[dict] =  field(default_factory=dict)
    develop: Optional[dict] = field(default_factory=dict)

    def __post_init__(self):
        """Run validation methods if declared.
        The validation method can be a simple check
        that raises ValueError or a transformation to
        the field value.
        The validation is performed by calling a function named:
            `validate_<field_name>(self, value) -> field.type`
        """
        super().__post_init__()
        self.meta = self._meta

    def validate__meta(self, value):
        return self.validate_meta(value)

    def validate_meta(self, value):
        if "_meta" in value:
            value = value["_meta"]
        if 'pipfile-spec' in value:
            value['pipfile_spec'] = value.pop('pipfile-spec')
        return Meta(**value)

    def validate_default(self, value):
        packages = {}
        for name, spec in value.items():
            packages[name] = Package(spec)
        return packages

    @classmethod
    def load(cls, fh, encoding=None):
        if encoding is None:
            data = json.load(fh)
        else:
            data = json.loads(fh.read().decode(encoding))
        return cls(**data)

    @classmethod
    def with_meta_from(cls, pipfile, categories=None):
        data = {
            "_meta": {
                "hash": pipfile.get_hash().__dict__,
                "pipfile-spec": PIPFILE_SPEC_CURRENT,
                "requires": _copy_jsonsafe(getattr(pipfile, "requires", {})),
            },
        }

        data["_meta"].update(asdict(pipfile.sources))

        if categories is None:
            data["default"] = _copy_jsonsafe(getattr(pipfile, "packages", {}))
            data["develop"] = _copy_jsonsafe(getattr(pipfile, "dev-packages", {}))
        else:
            for category in categories:
                if category in ["default", "packages"]:
                    data["default"] = _copy_jsonsafe(getattr(pipfile,"packages", {}))
                elif category in ["develop", "dev-packages"]:
                    data["develop"] = _copy_jsonsafe(
                            getattr(pipfile,"dev-packages", {}))
                else:
                    data[category] = _copy_jsonsafe(getattr(pipfile, category, {}))
        if "default" not in data:
            data["default"] = {}
        if "develop" not in data:
            data["develop"] = {}
        return cls(data)

    def __getitem__(self, key):
        value = self[key]
        try:
            if key == "_meta":
                return Meta(**value)
            return PackageCollection(value)
        except KeyError:
            return value

    def is_up_to_date(self, pipfile):
        return self.meta.hash == pipfile.get_hash()

    def dump(self, fh):
        json.dump(self, fh, cls=DCJSONEncoder)
        self.meta = self._meta

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value



import hashlib

import pytest

from plette import models
from plette.models import Source, SourceCollection


@pytest.fixture()
def sources():
    return models.SourceCollection(
        [
            {
                "name": "pypi",
                "url": "https://pypi.org/simple",
                "verify_ssl": True,
            },
            {
                "name": "devpi",
                "url": "http://127.0.0.1:$DEVPI_PORT/simple",
                "verify_ssl": True,
            },
        ]
    )


def test_get_slice(sources):
    sliced = sources[:1]
    assert isinstance(sliced, models.SourceCollection)
    assert len(sliced) == 1
    assert sliced[0] == models.Source(
        **{
            "name": "pypi",
            "url": "https://pypi.org/simple",
            "verify_ssl": True,
        }
    )


def test_set_slice(sources):
    sources[1:] = [
        Source(**{
            "name": "localpi-4433",
            "url": "https://127.0.0.1:4433/simple",
            "verify_ssl": False,
        }),
        Source(**{
            "name": "localpi-8000",
            "url": "http://127.0.0.1:8000/simple",
            "verify_ssl": True,
        }),
    ]
    assert sources == \
        SourceCollection([
        Source(**{
            "name": "pypi",
            "url": "https://pypi.org/simple",
            "verify_ssl": True,
        }),
        Source(**{
            "name": "localpi-4433",
            "url": "https://127.0.0.1:4433/simple",
            "verify_ssl": False,
        }),
        Source(**{
            "name": "localpi-8000",
            "url": "http://127.0.0.1:8000/simple",
            "verify_ssl": True,
        }),
        ])


def test_del_slice(sources):
    del sources[:1]
    assert sources == SourceCollection([
        Source(**{
            "name": "devpi",
            "url": "http://127.0.0.1:$DEVPI_PORT/simple",
            "verify_ssl": True,
        }),
    ])



# pylint: disable=missing-module-docstring,missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=no-member
# pylint: disable=too-few-public-methods
import os
import re
import shlex


from dataclasses import dataclass

from typing import Optional, List, Union


class ValidationError(ValueError):
    pass


def remove_empty_values(d):
    #  Iterate over a copy of the dictionary
    for key, value in list(d.items()):
        # If the value is a dictionary, call the function recursively
        if isinstance(value, dict):
            remove_empty_values(value)
            # If the dictionary is empty, remove the key
            if not value:
                del d[key]
        # If the value is None or an empty string, remove the key
        elif value is None or value == '':
            del d[key]


class BaseModel:

    def __post_init__(self):
        """Run validation methods if declared.
        The validation method can be a simple check
        that raises ValueError or a transformation to
        the field value.
        The validation is performed by calling a function named:
            `validate_<field_name>(self, value) -> field.type`
        """
        for name, _ in self.__dataclass_fields__.items():
            if (method := getattr(self, f"validate_{name}", None)):
                setattr(self, name, method(getattr(self, name)))


@dataclass
class Hash(BaseModel):

    name: str
    value: str

    def validate_name(self, value):
        if not isinstance(value, str):
            raise ValueError("Hash.name must be a string")

        return value

    def validate_value(self, value):
        if not isinstance(value, str):
            raise ValueError("Hash.value must be a string")

        return value

    @classmethod
    def from_hash(cls, ins):
        """Interpolation to the hash result of `hashlib`.
        """
        return cls(name=ins.name, value=ins.hexdigest())

    @classmethod
    def from_dict(cls, value):
        """parse a depedency line and create an Hash object"""
        try:
            name, value = list(value.items())[0]
        except AttributeError:
            name, value = value.split(":", 1)
        return cls(name, value)

    @classmethod
    def from_line(cls, value):
        """parse a dependecy line and create a Hash object"""
        try:
            name, value = value.split(":", 1)
        except AttributeError:
            name, value = list(value.items())[0]
        return cls(name, value)

    def __eq__(self, other):
        if not isinstance(other, Hash):
            raise TypeError(f"cannot compare Hash with {type(other).__name__!r}")
        return self.value == other.value

    def as_line(self):
        return f"{self.name}:{self.value}"


@dataclass
class Source(BaseModel):
    """Information on a "simple" Python package index.

    This could be PyPI, or a self-hosted index server, etc. The server
    specified by the `url` attribute is expected to provide the "simple"
    package API.
    """
    name: str
    verify_ssl: bool
    url: str

    @property
    def url_expanded(self):
        return os.path.expandvars(self.url)

    def validate_verify_ssl(self, value):
        if not isinstance(value, bool):
            raise ValidationError("verify_ssl: must be of boolean type")
        return value


@dataclass
class PackageSpecfiers(BaseModel):

    extras: List[str]

    def validate_extras(self, value):
        if not isinstance(value, list):
            raise ValidationError("Extras must be a list")


@dataclass
class Package(BaseModel):

    version: Union[Optional[str],Optional[dict]] = "*"
    specifiers: Optional[PackageSpecfiers] = None
    editable: Optional[bool] = None
    extras: Optional[PackageSpecfiers] = None
    path: Optional[str] = None

    def validate_extras(self, value):
        if value is None:
            return value
        if not (isinstance(value, list) and all(isinstance(i, str) for i in value)):
            raise ValidationError("Extras must be a list or None")
        return value

    def validate_version(self, value):
        if isinstance(value, dict):
            return value
        if isinstance(value, str):
            return value
        if value is None:
            return "*"

        raise ValidationError(f"Unknown type {type(value)} for version")


@dataclass(init=False)
class Script(BaseModel):

    script:  Union[str, List[str]]

    def __init__(self, script):

        if isinstance(script, str):
            script = shlex.split(script)
        self._parts = [script[0]]
        self._parts.extend(script[1:])

    def validate_script(self, value):
        if not (isinstance(value, str) or
                (isinstance(value, list) and all(isinstance(i, str) for i in value))
                ):
            raise ValueError("script must be a string or a list of strings")

    def __repr__(self):
        return f"Script({self._parts!r})"

    @property
    def command(self):
        return self._parts[0]

    @property
    def args(self):
        return self._parts[1:]

    def cmdify(self, extra_args=None):
        """Encode into a cmd-executable string.

        This re-implements CreateProcess's quoting logic to turn a list of
        arguments into one single string for the shell to interpret.

        * All double quotes are escaped with a backslash.
        * Existing backslashes before a quote are doubled, so they are all
          escaped properly.
        * Backslashes elsewhere are left as-is; cmd will interpret them
          literally.

        The result is then quoted into a pair of double quotes to be grouped.

        An argument is intentionally not quoted if it does not contain
        whitespaces. This is done to be compatible with Windows built-in
        commands that don't work well with quotes, e.g. everything with `echo`,
        and DOS-style (forward slash) switches.

        The intended use of this function is to pre-process an argument list
        before passing it into ``subprocess.Popen(..., shell=True)``.

        See also: https://docs.python.org/3/library/subprocess.html
        """
        parts = list(self._parts)
        if extra_args:
            parts.extend(extra_args)
        return " ".join(
            arg if not next(re.finditer(r'\s', arg), None)
            else '"{0}"'.format(re.sub(r'(\\*)"', r'\1\1\\"', arg))
            for arg in parts
        )


@dataclass
class PackageCollection(BaseModel):

    packages: List[Package]

    def validate_packages(self, value):
        if isinstance(value, dict):
            packages = {}
            for k, v in value.items():
                if isinstance(v, dict):
                    packages[k] = Package(**v)
                else:
                    packages[k] = Package(version=v)
            return packages
        return value


@dataclass
class ScriptCollection(BaseModel):
    scripts: List[Script]


@dataclass
class SourceCollection(BaseModel):

    sources: List[Source]

    def validate_sources(self, value):
        sources = []
        for v in value:
            if isinstance(v, dict):
                sources.append(Source(**v))
            elif isinstance(v, Source):
                sources.append(v)
        return sources

    def __iter__(self):
        return (d for d in self.sources)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return SourceCollection(self.sources[key])
        if isinstance(key, int):
            src = self.sources[key]
            if isinstance(src, dict):
                return Source(**key)
            if isinstance(src, Source):
                return src
        raise TypeError(f"Unextepcted type {type(src)}")

    def __len__(self):
        return len(self.sources)

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            self.sources[key] = value
        elif isinstance(value, Source):
            self.sources.append(value)
        elif isinstance(value, list):
            self.sources.extend(value)
        else:
            raise TypeError(f"Unextepcted type {type(value)} for {value}")

    def __delitem__(self, key):
        del self.sources[key]


@dataclass
class Requires(BaseModel):
    python_version: Optional[str] = None
    python_full_version: Optional[str] = None


META_SECTIONS = {
    "hash": Hash,
    "requires": Requires,
    "sources": SourceCollection,
}


@dataclass
class PipfileSection(BaseModel):

    """
    Dummy pipfile validator that needs to be completed in a future PR
    Hint: many pipfile features are undocumented in pipenv/project.py
    """


@dataclass
class Meta(BaseModel):

    hash: Hash
    pipfile_spec: str
    requires: Requires
    sources: SourceCollection

    @classmethod
    def from_dict(cls, d: dict) -> "Meta":
        return cls(**{k.replace('-', '_'): v for k, v in d.items()})

    def validate_hash(self, value):
        try:
            return Hash(**value)
        except TypeError:
            return Hash.from_line(value)

    def validate_requires(self, value):
        return Requires(value)

    def validate_sources(self, value):
        return SourceCollection(value)

    def validate_pipfile_spec(self, value):
        if int(value) != 6:
            raise ValueError('Only pipefile-spec version 6 is supported')
        return value


@dataclass
class Pipenv(BaseModel):
    """Represent the [pipenv] section in Pipfile"""
    allow_prereleases: Optional[bool] = False
    install_search_all_sources: Optional[bool] = True

    def validate_allow_prereleases(self, value):
        if not isinstance(value, bool):
            raise ValidationError('allow_prereleases must be a boolean')
        return value

    def validate_install_search_all_sources(self, value):
        if not isinstance(value, bool):
            raise ValidationError('install_search_all_sources must be a boolean')

        return value



import pytest
from plette import models

def test_requires_python_version():
    r = models.Requires(**{"python_version": "8.19"})
    assert r.python_version == "8.19"


def test_requires_python_version_no_full_version():
    r = models.Requires(**{"python_version": "8.19"})
    r.python_full_version is None


def test_requires_python_full_version():
    r = models.Requires(**{"python_full_version": "8.19"})
    assert r.python_full_version == "8.19"


def test_requires_python_full_version_no_version():
    r = models.Requires(**{"python_full_version": "8.19"})
    r.python_version is None


def test_allows_python_version_and_full():
    r = models.Requires(**{"python_version": "8.1", "python_full_version": "8.1.9"})
    assert r.python_version == "8.1"
    assert r.python_full_version == "8.1.9"



from plette.models import Meta

HASH = "9aaf3dbaf8c4df3accd4606eb2275d3b91c9db41be4fd5a97ecc95d79a12cfe6"

def test_meta():
    m = Meta.from_dict(
        {
            "hash": {"sha256": HASH},
            "pipfile-spec": 6,
            "requires": {},
            "sources": [
                {
                    "name": "pypi",
                    "url": "https://pypi.org/simple",
                    "verify_ssl": True,
                },
            ],
          }
    )
    assert m.hash.name == "sha256"



import pytest

from plette.models import Package

def test_package_str():
    p = Package("*")
    assert p.version == "*"


def test_package_dict():
    p = Package({"version": "*"})
    assert p.version == {"version": "*"}


def test_package_version_is_none():
    p = Package(**{"path": ".", "editable": True})
    assert p.version == "*"
    assert p.editable is True

def test_package_with_wrong_extras():
    with pytest.raises(ValueError):
        p = Package(**{"version": "==1.20.0", "extras": "broker"})

    with pytest.raises(ValueError):
        p = Package(**{"version": "==1.20.0", "extras": ["broker", {}]})

    with pytest.raises(ValueError):
        p = Package(**{"version": "==1.20.0", "extras": ["broker", 1]})


def test_package_with_extras():
    p = Package(**{"version": "==1.20.0", "extras": ["broker", "tests"]})
    assert p.extras == ['broker', 'tests']


def test_package_wrong_key():
    p = Package(**{"path": ".", "editable": True})
    assert p.editable is True
    assert p.version is "*"



import hashlib

from plette.models import Hash

def test_hash_from_hash():
    v = hashlib.md5(b"foo")
    h = Hash.from_hash(v)
    assert h.name == "md5"
    assert h.value == "acbd18db4cc2f85cedef654fccc4a4d8"


def test_hash_from_line():
    h = Hash.from_line("md5:acbd18db4cc2f85cedef654fccc4a4d8")
    assert h.name == "md5"
    assert h.value == "acbd18db4cc2f85cedef654fccc4a4d8"


def test_hash_as_line():
    h = Hash.from_dict({"md5": "acbd18db4cc2f85cedef654fccc4a4d8"})
    assert h.as_line() == "md5:acbd18db4cc2f85cedef654fccc4a4d8"
